/*
 * CSE 154 Summer 2019
 * API for Fake Piazza
 *
 * An API that can log users in to a fake Piazza service and provide statistics on their
 * fake usage of the service.
 * *** Endpoint Documentation ***
 * Endpoint: /login
 * Description: Logs the user into the service with given credentials.
 * Request Type: POST
 * Response Type: Plain Text
 * Required Parameters: username, password
 * Example Request: /login {username: "Dubs", password: "boundless"}
 * Example Response: Successfully logged in!
 * Error Handling: Responds with 400 plain text if they do not give required params or they
 *                 provide invalid credentials.
 * *************************************
 * Endpoint: /logout
 * Description: Logs a user out if they are logged in.
 * Request Type: POST
 * Response Type: Plain Text
 * Example Request: /logout
 * Example Response: Successfully logged out!
 * *************************************
 * Endpoint: /stats
 * Description: Provides usage statistics for the logged in individual, if logged in.
 * Request Type: GET
 * Response Type: JSON
 * Example Request: /stats
 * Example Response:
 * {
 *   "Username": "Dubs",
 *   "Questions Asked": 20,
 *   "Courses": [
 *                "CSE 154",
 *                "MATH 126"
 *              ]
 *   "Responses Recieved": 11
 * }
 * Error Handling: Responds in text with 400 if the user is not logged in.
 */

"use strict";

const express = require("express");
const mysql = require("promise-mysql");
const multer = require("multer"); // Handles form-data requests.
const cookieParser = require("cookie-parser");
const app = express();

app.use(express.urlencoded({extended: true}));
app.use(express.json());
app.use(multer().none());
app.use(cookieParser());

const INVALID_PARAM_ERROR = 400;
const SERVER_ERROR = 500;
const SERVER_ERROR_MSG = "Something went wrong on the server.";
const COOKIE_EXPIRATION = 1000 * 60 * 60 * 3;

// Logs the user into the webservice, setting a login cookie that expires in 3 hours.
app.post("/login", async function(req, res) {
  res.type("text");
  let user = req.body.username;
  let pass = req.body.password;
  if (!user || !pass) {
    res.status(INVALID_PARAM_ERROR).send("Missing required parameters username and password!");
  } else {
    let db;
    try {
      db = await getDB();
      if (!(await checkCredentials(user, pass, db))) {
        res.status(INVALID_PARAM_ERROR).send("Invalid credentials.");
      } else {
        let id = await getSessionId(db);
        await setSessionId(id, user, db);
        db.end();
        res.cookie("sessionid", id, {expires: new Date(Date.now() + COOKIE_EXPIRATION)});
        res.send("Successfully logged in!");
      }
    } catch (err) {
      if (db) {
        db.end();
      }
      res.status(SERVER_ERROR).send(SERVER_ERROR_MSG);
    }
  }
});

// Logs a user out by expiring their cookie.
app.post("/logout", function(req, res) {
  res.type("text");
  let id = req.cookies["sessionid"];
  if (id) {
    res.clearCookie("sessionid");
    res.send("Successfully logged out!");
  } else {
    res.send("Already logged out.");
  }
});

app.get("/stats", async function(req, res) {
  let id = req.cookies["sessionid"];
  if (!id) {
    res.type("text");
    res.status(INVALID_PARAM_ERROR).send("You are not logged in");
  } else {
    let db;
    try {
      db = await getDB();
      let name = await checkSession(id, db);
      if (!name) {
        res.type("text");
        res.send("Invalid login token, please log in again!");
      } else {
        let questions = await getQuestionsAsked(id, db);
        let courses = await getCoursesTaken(id, db);
        let responses = await getResponsesRecieved(id, db);
        res.json(makeStats(name, questions, courses, responses));
      }
    } catch (err) {
      if (db) {
        db.end();
      }
      res.type("text");
      res.status(SERVER_ERROR).send(SERVER_ERROR_MSG);
    }
  }
});

/**
 * Establishes a database connection and returns the database object.
 * @returns {Object} - The database object for the connection.
 */
async function getDB() {
  let db = await mysql.createConnection({
    host: "localhost",
    port: "3307",
    user: "root",
    password: "root",
    database: "piazzadb"
  });
  return db;
}

/**
 * Generates an unused sessionid and returns it to the user.
 * @param {object} db - A connection object to the database.
 * @returns {string} - The random session id.
 */
async function getSessionId(db) {
  let query = "SELECT sessionid FROM users WHERE sessionid = ?";
  let id;
  do {
    // This wizardry comes from https://gist.github.com/6174/6062387
    id = Math.random().toString(36)
      .substring(2, 15) + Math.random().toString(36)
      .substring(2, 15);
  } while ((await db.query(query, [id])).length > 0);
  return id;
}

/**
 * @param {string} user - The username to check
 * @param {string} pass - The password to check
 * @param {object} db - The database connection object.
 * @returns {boolean} - True if the credentials match for a user, false otherwise.
 */
async function checkCredentials(user, pass, db) {
  let query = "SELECT username FROM users WHERE username = ? AND password = ?;";
  return ((await db.query(query, [user, pass])).length > 0);
}

/**
 * Sets the session id in the database to the given one for the given user.
 * @param {string} id - The Session id to set
 * @param {string} user - The username of the person to set the id for
 * @param {object} db - The database connection object
 */
async function setSessionId(id, user, db) {
  let query = "UPDATE users SET sessionid = ? WHERE username = ?";
  await db.query(query, [id, user]);
}

/**
 * Returns the username if the sessionid is valid, or null if it is invalid.
 * @param {string} id - The sessionid to check
 * @param {object} db - The database connection object
 * @returns {string} - The username corresponding to the id, or null if there is none.
 */
async function checkSession(id, db) {
  let query = "SELECT username FROM users WHERE sessionid = ?";
  let results = await db.query(query, [id]);
  if (results.length > 0) {
    return results[0]["username"];
  } else {
    return null;
  }
}

/**
 * Constructs the final JSON result for the /stats endpoint.
 * @param {string} name - The username of the user.
 * @param {number} questions - The number of questions the user asked
 * @param {array} courses - The list of courses taken by the user
 * @param {number} responses - The number of responses the user recieved on their questions.
 * @returns {object} - The JSON object to send to the client.
 */
function makeStats(name, questions, courses, responses) {
  let stats = {username: name};
  stats["Questions Asked"] = questions;
  stats["Courses"] = courses;
  stats["Responses Recieved"] = responses;
  return stats;
}

/**
 * Returns the number of questions asked by the user with the given sessionid
 * @param {string} id - The sessionid of the user
 * @param {object} db - The database connection object
 * @returns {number} - The number of questions asked by the user
 */
async function getQuestionsAsked(id, db) {
  let query = "SELECT q.id FROM questions q JOIN users u ON u.id = q.asker WHERE u.sessionid = ?";
  let results = await db.query(query, [id]);
  return results.length;
}

/**
 * Returns an array of course names taken by the user with the given sessionid
 * @param {string} id - the sessionid of the user
 * @param {object} db - The database connection object
 * @returns {number} - The number of questions asked by the user
 */
async function getCoursesTaken(id, db) {
  let query = "SELECT c.name FROM courses c JOIN enrollments e ON e.course = c.id " +
              "JOIN users u ON u.id = e.user WHERE u.sessionid = ?";
  let results = await db.query(query, [id]);
  let courses = [];
  for (let i = 0; i < results.length; i++) {
    courses.push(results[i]["name"]);
  }
  return courses;
}

/**
 * Returns the number of responses recieved on questions by the user with the given sessionid
 * @param {string} id - The sessionid of the user
 * @param {object} db - The database connection object
 * @returns {number} - The number of responses recieved
 */
async function getResponsesRecieved(id, db) {
  let query = "SELECT r.id FROM responses r JOIN questions q ON q.id = r.question " +
              "JOIN users u ON u.id = q.asker WHERE u.sessionid = ?";
  let results = await db.query(query, [id]);
  return results.length;
}

const PORT = process.env.PORT || 8000;
app.listen(PORT);
